/*
 * Copyright (C) 2012 Wayne Meissner
 *
 * This file is part of the JNR project.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package jnr.ffi;

public enum TypeAlias {
    int8_t,
    u_int8_t,
    int16_t,
    u_int16_t,
    int32_t,
    u_int32_t,
    int64_t,
    u_int64_t,
    intptr_t,
    uintptr_t,
    caddr_t,
    dev_t,
    blkcnt_t,
    blksize_t,
    gid_t,
    in_addr_t,
    in_port_t,
    ino_t,
    ino64_t,
    key_t,
    mode_t,
    nlink_t,
    id_t,
    pid_t,
    off_t,
    swblk_t,
    uid_t,
    clock_t,
    size_t,
    ssize_t,
    time_t,
    fsblkcnt_t,
    fsfilcnt_t,
    sa_family_t,
    socklen_t,
    rlim_t,
}
